$       TT("XA eForth")

$       PL(60)          ;60 lines per page
$       PW(120)         ;120 characters per line
$       GEN             ;expand macros
;===============================================================
;XA eForth 2.07 by C. H. Ting, 1998
;
;1.     Add interrupt code to drive S0 serial port.
;       Use two 256 circular queues for receiving and transmitting.
;       4 byte queue pointers are defined at the end of user area.
;       Interupts and queues are checked out. 27mar98cht v2.02
;2.     Add Timer 2 control words T2START and T2DONE. v2.03
;3.     Add PAUSE, and Task1. Delete DIAGNOSE. V2.04
;       Task switching works, and it is fast! 10apr98cht
;4.     Fix task dictionary pointers. 14apr98cht
;       Plus24.txt loaded and seems ok. 
;       Delete 8051 port code.  Delete FILE, HAND, CONSOLE.
;       Move task1 to FD00-FFFF.
;5.     Host task is also interrupt driven, 15apr98cht
;       General clean-up.
;6.     Put PAUSE in ?KEY from KEY, 24apr98cht
;       Add SEND and FORGET, 27apr98cht
;
;XA eForth 2.01 by C. H. Ting, 1998
;
;1.     redesign of the eForth model for this Harvard archtecture.
;       Dictionary is in the ROM code space, but copied to RAM at boot.
;       Code words are executed in ROM.  Colon words are executed in
;       RAM space, and can be extended in the RAM space.
;
;       RAM memory configuration:
;       0000-00FF       User variables and all other variables
;       0100-017F       TIB and return stack
;       0180-01FF       Data stack
;       0200-03ff       Do not use, on-board RAM, may not exist
;       0400-07ff       On board i/o devices
;       1000-2800       Code dictionary
;       2000-6FFF       Dictionary expansion space
;       7000-7FFF       Name dictionary
;
;       ROM memory configuartion:
;       0000-011F       Interrupt vectors
;       0120            Interrupt service, default
;       0180-01FF       Initial user variables, copied to RAM
;       0200-           Cold start
;       1000-7FFF       Initial dictionary, copied to RAM
;
;2.     Indirect thread model to allow colon words to be executed
;       in the data RAM space. Use R2 as WP.
;       Code space and data space cannot be mixed in our hardware.
;       Code words must be executed in the code space.
;       Colon words has to be in the data space, to be extensible.
;       Entire dictionary is copied from code space to data space,
;               keeping addresses unchanged.
;       Code words are still executed in ROM, but data in them are
;               accessible because an identical copy is in RAM.
;       While running, no new code word can be defined.  However,
;               colon words, variables and constants can be added freely.
;3.     Assembled in Medium Model, 64K of RAM and 64K of ROM, in HPDXA
;               from Avocet.
;4.     Passing Ken Chen's diagnose. 01/07/98 cht
;5.     Fixed bugs and brought up interpreter. 01/12/98
;       New definitions do not work because jmp [[r2+]] fetches
;       address in the code space, not in the data space as I had hoped.
;       I need a separate set of inner interpreters for user defined
;       words.
;6.     Code has to start above the internal RAM, from 1000H to avoid
;       memory collison.
;7.     Code dictionary starts at 1000H.  Name dictionary starts at
;       8000H growing downward.  
;       Compiler works. 01/13/98 cht
;       Change jmp [[r2+]] to mov.w r4,[r2+] jmp [r4]
;       so that new words will work with old words in data space.
;       Both code and colon words are copied to RAM.
;8.     Add multitasking mechanism, implementation is deferred.
;9.     Fully checked out on XA target board.  Code starts at 1000H.
;10.    Add square root of 32-bit unsigned integer. eforthx5.as
;       Access peripheral devices, and also drive stepper motor.
;
;XA eForth 1.01 by C. H. Ting, 1997
;
;       Modified for 8051 eForth.  Assembled with XA macro assembler
;       from Avocet Systems, Inc.
;1.     Fully optimized for the 16-bit 80C51XA microcontrollers from
;       Philips Semiconductors.  The most interesting thing about XA
;       is the instruction JMP [[R3+]], which is the Forth NEXT itself!
;
;       next:   jmp [[r3+]]
;       docol:  xch     r7,r3
;               next
;       exit:   pop     r3
;               next
;
;2.     Add DOVAR, DOCON, DOUSE, DOSTR, all headeless.
;3.     Delete headers for words not useful to general users.
;4.     All tool words reduced to code whenever possible.
;5.     Assume ROM between 0 and 7FFFH and RAM between 8000H and FFFFH.
;6.     This implementation was not implemented.  It was a mental exercise
;               as a prelude to the 2.01 implemetation.
;       80C51XA is a Harvard architecture machine, which is not quite
;               compatible to direct threaded Forth.  Serious trade-offs 
;               must be considered. Indirect treaded model is more
;               suitable.
;7.     This model is ok for a Forth executed entirely from a ROM.
;       It will not be extensible, because new definitions cannot be
;               added while running.  New word must be in the code
;               space, because code field contains executable code.
;               Adding new words in the data space, indirection is
;               required.

;8051 eForth 1.1 by C. H. Ting, 1990
;
;       This eForth system was developed using chipForth from Forth, Inc.
;       and tested on a Micromint BCC52 single board computer.
;       The eForth Model was developed by Bill Muench and C. H. Ting.
;
;
;       Direct your questions and contributions to:
;
;               Dr. C. H. Ting
;               156 14th Avenue
;               San Mateo, CA 94402
;               (650) 571-7639
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; The virtual machine
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Register assignments
;
;       R0      SP      data stack pointer
;       R1      TOS     top of data stack cached here
;       R2      WP      current word pointer
;       R3      IP      instruction pointer
;       R4      A       scratch pad 1
;       R5      B       scratch pad 2
;       R6      UP      user area pointer
;       R7      RP      return stack pointer
;
; Forth virtual machine
;
;;      next:   mov.w   r2,[r3+]        ;;get next addr from IP
;;              jmp     [[r2+]]         ;;make indirect jump to it
;       ;;WP now has pfa
;       next:   mov.w   r2,[r3+]        ;;execution addresses
;               mov.w   r4,[r2+]        ;;are in data space
;               jmp     [r4]            ;;can't use jmp [[r2+]]
;               
;
;       docol:  push    r3              ;;save IP on return stack
;               mov.w   r3,r2           ;;get new IP from WP
;               next    ;;go do the colon word list
;
;       exit:   pop     r3              ;;restore IP from return stack
;               next    ;;resume

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Constants and equates
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Version control

VER     EQU     02H     ;major release version
EXT     EQU     07H     ;minor extension

;; Constants

COMPO   EQU     040H    ;lexicon compile only bit
IMEDD   EQU     080H    ;lexicon immediate bit
MASKK   EQU     07F1FH  ;lexicon bit mask

CELLL   EQU     2       ;size of a cell
BASEE   EQU     10      ;default radix
VOCSS   EQU     8       ;depth of vocabulary stack

BKSPP   EQU     8       ;backspace
LF      EQU     10      ;line feed
CRR     EQU     13      ;carriage return
ERR     EQU     27      ;error escape
TIC     EQU     39      ;tick

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Memory allocation
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;  ROM map

EM      EQU     8000H   ;top of ROM dictionary
BM      EQU     0H      ;bottom of ROM dictionary
USERR   EQU     0140H   ;initial user variables
COLDD   EQU     0200H   ;cold start vector
CODEE   EQU     1000H   ;initial code dictionary
NAMEE   EQU     8000H   ;initial name dictionary

;  RAM map

UPP     EQU     0       ;start of 1st user area (UP0)
TIBB    EQU     050H    ;terminal input buffer (TIB)
RPP     EQU     0A8H    ;start of return stack (RP0)
SPP     EQU     0F8H    ;start of data stack (SP0)

UPP1    EQU     100H    ;start of 2nd user area (UP0)
TIBB1   EQU     150H    ;terminal input buffer (TIB)
RPP1    EQU     1A8H    ;start of return stack (RP0)
SPP1    EQU     1F8H    ;start of data stack (SP0)
CTOP1   EQU     0FF00H  ;code dictionary
NTOP1   EQU     0FFFEH  ;name dictionary

;  Register equates

BCR     EQU     46AH
BTRH    EQU     469H
BTRL    EQU     468H
EA      EQU     337H
ERI0    EQU     338H
ERI1    EQU     33AH    
IEH     EQU     427H
IPA4    EQU     4A4H
IPA5    EQU     4A5H
P0      EQU     430H
P1      EQU     431H
P2      EQU     432H
P3      EQU     433H
P0CFGA  EQU     470H
P1CFGA  EQU     471H
P2CFGA  EQU     472H
P3CFGA  EQU     473H
P0CFGB  EQU     4F0H
P1CFGB  EQU     4F1H
P2CFGB  EQU     4F2H
P3CFGB  EQU     4F3H
PSWH    EQU     401H
RS0     EQU     20CH
RTL0    EQU     454H
RTH0    EQU     455H
RTL1    EQU     456H
RTH1    EQU     457H
S0CON   EQU     420H
S0BUF   EQU     460H
S1CON   EQU     424H
S1BUF   EQU     464H
TL0     EQU     450H
TH0     EQU     451H
TL1     EQU     452H
TH1     EQU     453H
TL2     EQU     458H
TH2     EQU     459H
T2CAPL  EQU     45AH
T2CAPH  EQU     45BH
T2CON   EQU     418H
TF2     EQU     2C7H
TR0     EQU     284H    ;run timer 0
TR1     EQU     286H    ;run timer 1 for serial port 0
TR2     EQU     2C2H
WDCON   EQU     41FH
WFEED1  EQU     45DH
WFEED2  EQU     45EH

;; Initialize assembly variables

_LINK   SET     0       ;force a null link
_NAME   SET     NAMEE   ;initialize name pointer
_CFA    SET     CODEE   ;initialize code pointer
_USER   SET     0               ;first user variable offset

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Define assembly macros
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; The inner interpreter NEXT

next    macro   ;;Indirect thread model
        mov.w   r2,[r3+]
        mov.w   r4,[r2+]
        jmp     [r4]
        endm

; Stack macros

pushs   macro   ;;DUP
        adds.w  r0,#-2
        mov.w   [r0],r1
        endm

pops    macro   ;;DROP
        mov.w   r1,[r0+]
        endm

; Compile a definition header.

HEADER  MACRO   LEX,NAME,LABEL
        ALIGN.W ;;force to cell boundary
LABEL:  ;;assembly label
_CFA    SET     $       ;;save code pointer
_LEN    SET     (LEX AND 01FH)/CELLL    ;;string cell count, round down
_TEMP   SET     _NAME
_NAME   SET     _TEMP-((_LEN+3)*CELLL)  ;;new header on cell boundary
        ORG     _NAME   ;;set name pointer
        DW      _CFA,_LINK              ;;token pointer and link
_LINK   SET     $       ;;link points to a name string
        DB      LEX,NAME                ;;name string
        ALIGN.W
        ORG     _CFA    ;;restore code pointer
        ENDM

;       Compile a code definition

SCODE   MACRO   LEX,NAME,LABEL
        ALIGN.W ;;force to cell boundary
LABEL:  ;;assembly label
_CFA    SET     $       ;;save code pointer
_LEN    SET     (LEX AND 01FH)/CELLL    ;;string cell count, round down
_TEMP   SET     _NAME
_NAME   SET     _TEMP-((_LEN+3)*CELLL)  ;;new header on cell boundary
        ORG     _NAME   ;;set name pointer
        DW      _CFA,_LINK              ;;token pointer and link
_LINK   SET     $       ;;link points to a name string
        DB      LEX,NAME                ;;name string
        ALIGN.W
        ORG     _CFA    ;;restore code pointer
        DW      $+2     ;;indirect thread model
        ENDM

;       Compile a colon definition 

SCOLON  MACRO   LEX,NAME,LABEL
        ALIGN.W ;;force to cell boundary
LABEL:  ;;assembly label
_CFA    SET     $       ;;save code pointer
_LEN    SET     (LEX AND 01FH)/CELLL    ;;string cell count, round down
_TEMP   SET     _NAME
_NAME   SET     _TEMP-((_LEN+3)*CELLL)  ;;new header on cell boundary
        ORG     _NAME   ;;set name pointer
        DW      _CFA,_LINK              ;;token pointer and link
_LINK   SET     $       ;;link points to a name string
        DB      LEX,NAME                ;;name string
        ALIGN.W
        ORG     _CFA    ;;restore code pointer
        DW      DOLST
        ENDM

;       Compile a user variable

SUSER   MACRO   LEX,NAME,LABEL
        ALIGN.W ;;force to cell boundary
LABEL:  ;;assembly label
_CFA    SET     $       ;;save code pointer
_LEN    SET     (LEX AND 01FH)/CELLL    ;;string cell count, round down
_TEMP   SET     _NAME
_NAME   SET     _TEMP-((_LEN+3)*CELLL)  ;;new header on cell boundary
        ORG     _NAME   ;;set name pointer
        DW      _CFA,_LINK              ;;token pointer and link
_LINK   SET     $       ;;link points to a name string
        DB      LEX,NAME                ;;name string
        ALIGN.W
        ORG     _CFA    ;;restore code pointer
        DW      DOUSE
        dw      _USER
        _USER   SET _USER+CELLL         ;;update user area offset
        ENDM

;       Compile an inline string.

DSS     MACRO   FUNCT,STRNG
        DW      FUNCT   ;;function
        _LEN    SET     $               ;;save address of count byte
        DB      0,STRNG ;;count byte and string
        _CFA    SET     $               ;;save code pointer
        ORG     _LEN    ;;point to count byte
        DB      _CFA-_LEN-1             ;;set count
        ORG     _CFA    ;;restore code pointer
        ALIGN.W
        ENDM


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Reset and interrupt vectors
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        ORG     BM
        DW      08000H,ORIG             ;RESET to cold start

        rept    47H     ;71 interrupt vectors
        DW      08000H,RETII            ;default RETI
        endm

RETII:  reti    ;default interrupt service

        ORG     0A0H
        DW      08A00H,R0INT
        ORG     0A8H
        DW      08900H,R1INT


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Initial user variables.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       ORG     CODEE
        ORG     USERR

UZERO:  
        DW      SPP     ;SP0
        DW      RPP     ;RP0
        DW      RX1     ;'?KEY
        DW      TXSTO   ;'EMIT
        DW      ACCEP   ;'EXPECT
        DW      KTAP    ;'TAP
        DW      TXSTO   ;'ECHO
        DW      DOTOK   ;'PROMPT
        DW      BASEE   ;BASE
        DW      0       ;tmp
        DW      0       ;SPAN
        DW      0       ;>IN
        DW      0       ;#TIB
        DW      TIBB    ;TIB
        DW      0       ;CSP
        DW      INTER   ;'EVAL
        DW      NUMBQ   ;'NUMBER
        DW      0       ;HLD
        DW      0       ;HANDLER
        DW      0       ;CONTEXT pointer
        DW      0,0,0,0,0,0,0,0 ;vocabulary stack
        DW      0       ;CURRENT pointer
        DW      0       ;vocabulary link pointer
        DW      CTOP    ;CP in RAM
        DW      NTOP    ;NP in RAM
        DW      LASTN   ;LAST
        DW      LASTN   ;FORTH
        DB      1       ;tHead,tTail,rHead,rTail
        DB      1
        DB      1
        DB      1       ;need non-zero's for HEX file
ULAST:
UZERO1:
        DW      SPP1    ;SP0
        DW      RPP1    ;RP0
        DW      RX0     ;'?KEY
        DW      TXSTO0  ;'EMIT
        DW      ACCEP   ;'EXPECT
        DW      KTAP    ;'TAP
        DW      TXSTO0  ;'ECHO
        DW      DOTOK   ;'PROMPT
        DW      BASEE   ;BASE
        DW      0       ;tmp
        DW      0       ;SPAN
        DW      0       ;>IN
        DW      0       ;#TIB
        DW      TIBB1   ;TIB
        DW      0       ;CSP
        DW      INTER   ;'EVAL
        DW      NUMBQ   ;'NUMBER
        DW      0       ;HLD
        DW      0       ;HANDLER
        DW      0       ;CONTEXT pointer
        DW      0,0,0,0,0,0,0,0 ;vocabulary stack
        DW      0       ;CURRENT pointer
        DW      0       ;vocabulary link pointer
        DW      CTOP1   ;CP in RAM
        DW      NTOP1   ;NP in RAM
        DW      LASTN   ;LAST
        DW      LASTN   ;FORTH
        DB      1       ;tHead,tTail,rHead,rTail
        DB      1
        DB      1
        DB      1       ;need non-zero's for HEX file
ULAST1:


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Cold boot
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        ORG     COLDD
        PUBLIC  _main

_main:
ORIG:   ;Cold boot routine
        movs.b  WDCON,#0                ;turn off watchdog timer
        mov.b   WFEED1,#0A5H            ;feed watchdog part 1
        mov.b   WFEED2,#05AH            ;feed watchdog part 2
        movs.b  BCR,#06H                ;16 bit data bus, 20 bit address
        mov.b   BTRH,#0FFH              ;slowest bus speed
        mov.b   BTRL,#0EFH
        mov.b   T2CON,#0                ;init Timer 2

        mov.w   r0,#50H
        mov.w   r1,#UZERO               ;from ROM
        mov.w   r2,#UPP ;to RAM
        call    copy    ;copy 1st user area

        mov.w   r0,#50H
        mov.w   r1,#UZERO1              ;from ROM
        mov.w   r2,#UPP1                ;to RAM
        call    copy    ;copy 1st user area

        mov.w   r0,#7000H               ;range 800-1FFF
        mov.w   r1,#CODEE               ;from ROM memory
        mov.w   r2,#CODEE               ;to RAM memory
        call    copy    ;copy dictionary

        mov.w   r7,#RPP ;init return stack
        movs.w  r6,#UPP ;init user area

        setb    RS0     ;register group 1
        mov.w   r0,#SPP1                ;init data stack SP1
        movs.w  r1,#0   ;init TOS1
        mov.w   r2,#RPP1                ;init RP1
        mov.w   r3,#COLD1               ;init IP1

        clr     RS0     ;register group 0
        mov.w   r0,#SPP ;init data stack
        movs.w  r1,#0   ;init SP
        mov.w   r3,#COLD1               ;init TOS

        MOV.B   S0CON,#52H              ;8 bit UART, enable Tx
        MOV.B   S1CON,#52H              ;8 bit UART, enable Tx
        MOV.B   R4L,#0E6h               ;9600 baud from timer 1
        MOV.B   TL1,R4L  
        MOV.B   RTL1,R4L
        MOV.B   R4L,#0FFH
        MOV.B   TH1,R4L
        MOV.B   RTH1,R4L
        SETB    TR1     ;run timer 1
        MOV.B   S0BUF,#0DH
        MOV.B   S1BUF,#0DH
        MOV.B   IPA5,#1

        mov.b   IPA4,#2H
        setb    ERI0    ;enable serial port 0 interrupts
        setb    ERI1
        setb    EA
        next    ;go Forth

        align.w
copy:   cmp     r0,#0   ;copy ROM data/code to RAM
        beq     copy2
copy1:  movc    r3l,[r1+]               ;get source byte
        mov     [r2+],r3l               ;store to destination
        djnz    r0,copy1
        align.w
copy2:  ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Interrupt driven serial i/o thru S0 port
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       Circular queues

rHead   EQU     44H     ;after all users variables
rTail   EQU     45H
rHead0  EQU     144H
rTail0  EQU     145H

rBuf    EQU     0FE00H
rBuf0   EQU     0FD00H

;
;       Receiver 0 Interrupt
;
        ALIGN.W

R0INT:  CLR     S0CON.0
        PUSH    R4
        MOV.B   R4L,rHead0
        ADDS.B  R4L,#1
        CMP.B   R4L,rTail0
        BEQ     R0IN1
        MOV.B   R4H,#(rBuf0/256)
        ADDS.B  R4L,#-1
        MOV.B   [R4],S0BUF
        ADDS.B  rHead0,#1
        NOP
R0IN1:  POP     R4
        RETI
;
;       Receiver 1 interrupt
;
        ALIGN.W

R1INT:  CLR     S1CON.0
        PUSH    R4
        MOV.B   R4L,rHead
        ADDS.B  R4L,#1
        CMP.B   R4L,rTail
        BEQ     R1IN1
        MOV.B   R4H,#(rBuf/256)
        ADDS.B  R4L,#-1
        MOV.B   [R4],S1BUF
        ADDS.B  rHead,#1
        NOP
R1IN1:  POP     R4
        RETI

;
;       ?RX0    ( -- c t| f)
;       Get character from receiver 0 queue
;
        SCODE   4,'?RX0',RX0
        MOV.B   R4L,rTail0
        CMP.B   R4L,rHead0
        BEQ     RX01
        pushs
        MOV.B   R4H,#(rBuf0/256)
        MOV.B   R1L,[R4]
        MOV.B   R1H,#0
        ADDS.B  rTail0,#1
        pushs
        MOV.W   R1,#-1
        next
        NOP
RX01:   pushs
        MOV.W   R1,#0
        next
;
;       ?RX1    ( -- c t| f )
;       Get character from receiver 1 queue
;
        SCODE   4,'?RX1',RX1
        MOV.B   R4L,rTail
        CMP.B   R4L,rHead
        BEQ     RX11
        pushs
        MOV.B   R4H,#(rBuf/256)
        MOV.B   R1L,[R4]
        MOV.B   R1H,#0
        ADDS.B  rTail,#1
        pushs
        MOV.W   R1,#-1
        next
        NOP
RX11:   pushs
        MOV.W   R1,#0
        next

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; The kernel    we add inner interpreters for constants, variables,
;               user variables, literals, and strings.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


        ORG     CODEE

;       doLIST  ( -- )
;       Process colon list.

        HEADER  046H,'doList',DOLST
        push    r3
        mov.w   r3,r2
        next

;       doCON   ( -- n )
;       Return value of constant in the parameter field.

        HEADER  045H,'doCon',DOCON
        pushs
        mov.w   r1,[r2]
        next

;       doVAR   ( -- a )
;       Return the parameter field address of variable.

        HEADER  045H,'doVar',DOVAR
        pushs
        mov.w   r1,r2
        next

;       doUSE   ( -- a )
;       Return the address of user variable in user area.
;       In XA, user area starts at 0 in RAM.  Offset is address.

        HEADER  045H,'doUse',DOUSE
        pushs
        mov.w   r1,[r2]
        add.w   r1,r6
        next

;       doLIT   ( -- w )
;       Push an inline literal.

        SCODE   045H,'doLit',DOLIT
        pushs
        mov.w   r1,[r3+]
        next

;       doSTR   ( -- a )
;       Return the address of the following string.

        SCODE   045H,'doStr',DOSTR
        pushs
        mov.w   r1,[r7] ;string addr is on return stack
        movs.w  r4,#0
        mov.b   r4l,[r1]                ;get count byte
        add.w   [r7],r4
        adds.w  [r7],#2 ;align return address
        and.w   [r7],#-2
        next

;       next    ( -- )
;       Run time code for the single index loop.
;       : next ( -- ) \ hilevel model
;       r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

        SCODE   044H,'next',DONXT
        sub.w   [r7],#1
        bcs     nxt1
        mov.w   r3,[r3]
        next
nxt1:   mov.w   [r7+],[r3+]             ;increment both RP and IP
        next    ;only 2 bytes and 5 cycles

;       ?branch ( f -- )
;       Branch if flag is zero.

        SCODE   047H,'?branch',QBRAN
        mov.w   r1,r1
        beq     QBRAN1
        adds.w  r3,#2
        pops
        next
QBRAN1: pops
        mov.w   r3,[r3]
        next

;       branch  ( -- )
;       Branch to an inline address.

        SCODE   046H,'branch',BRAN
        mov.w   r3,[r3]
        next

;       EXECUTE ( ca -- )
;       Execute the word at ca.

        SCODE   7,'EXECUTE',EXECU
        mov.w   r2,r1           
        pops
;       jmp     [[r2+]] ;WP must point to parameter field
        mov.w   r4,[r2+]
        jmp     [r4]

;       EXIT    ( -- )
;       Terminate a colon definition.

        SCODE   4,'EXIT',EXIT
        pop     r3
        next

;       !       ( w a -- )
;       Pop the data stack to memory.

        SCODE   1,021H,STORE
        mov.w   r4,[r0+]
        mov.w   [r1],r4
        pops
        next

;       @       ( a -- w )
;       Push memory location to the data stack.

        SCODE   1,'@',AT
        mov.w   r1,[r1]
        next

;       C!      ( c b -- )
;       Pop the data stack to byte memory.

        SCODE   2,'C!!',CSTOR
        mov.w   r4,[r0+]
        mov.b   [r1],r4l
        pops
        next

;       C@      ( b -- c )
;       Push byte memory location to the data stack.

        SCODE   2,'C@',CAT
        mov.b   r1l,[r1]
        movs.b  r1h,#0
        next

;       >R      ( w -- )
;       Push the data stack to the return stack.

        SCODE   2,'>R',TOR
        push    r1
        pops
        next

;       R@      ( -- w )
;       Copy top of return stack to the data stack.

        SCODE   2,'R@',RAT
        pushs
        mov.w   r1,[r7]
        next

;       R>      ( -- w )
;       Pop the return stack to the data stack.

        SCODE   2,'R>',RFROM
        pushs
        pop     r1
        next

;       RP@     ( -- a )
;       Push the current RP to the data stack.

        SCODE   3,'RP@',RPAT
        pushs
        mov.w   r1,r7
        next

;       RP!     ( a -- )
;       Set the return stack pointer.

        SCODE   043H,'RP!!',RPSTO
        mov.w   r7,r1
        pops
        next

;       SP@     ( -- a )
;       Push the current data stack pointer.

        SCODE   3,'SP@',SPAT
        pushs
        mov.w   r1,r0
        next

;       SP!     ( a -- )
;       Set the data stack pointer.

        SCODE   3,'SP!!',SPSTO
        mov.w   r0,r1
        movs.w  r1,#0   ;init TOS
        next

;       DUP     ( w -- w w )
;       Duplicate the top stack item.

        SCODE   3,'DUP',DUPP
        pushs
        next

;       DROP    ( w -- )
;       Discard top stack item.

        SCODE   4,'DROP',DROP
        pops
        next

;       SWAP    ( w1 w2 -- w2 w1 )
;       Exchange top two stack items.

        SCODE   4,'SWAP',SWAP
        xch.w   r1,[r0]
        next

;       OVER    ( w1 w2 -- w1 w2 w1 )
;       Copy second stack item to top.

        SCODE   4,'OVER',OVER
        pushs
        mov.w   r1,[r0+2]
        next

;       0<      ( n -- t )
;       Return true if n is negative.

        SCODE   2,'0!<',ZLESS
        mov.w   r1,r1
        sext.w  r1
        next

;       AND     ( w w -- w )
;       Bitwise AND.

        SCODE   3,'AND',ANDD
        and.w   r1,[r0+]
        next

;       OR      ( w w -- w )
;       Bitwise inclusive OR.

        SCODE   2,'OR',ORR
        or.w    r1,[r0+]
        next

;       XOR     ( w w -- w )
;       Bitwise exclusive OR.

        SCODE   3,'XOR',XORR
        xor.w   r1,[r0+]
        next

;       UM+     ( w w -- w cy )
;       Add two numbers, return the sum and carry flag.

        SCODE   3,'UM+',UPLUS
        add.w   [r0],r1
        bcc     uplus1
        movs.w  r1,#1
        next
uplus1: movs.w  r1,#0
        next


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Device dependent I/O
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;       TX!     ( c -- )
;       Send character c to the output device.

_TX:    ;for diagnose
        SCODE   3,'tx!!',TXSTO
txst0:  jnb     S1CON.1,txst0           ;txst0 should not be here
txst1:  clr     S1CON.1 ;debugging only
        mov.b   S1BUF,r1l
        pops
        next

;       TX0!    ( c -- )
;       Send character c to the output device.

        SCODE   4,'tx0!!',TXSTO0
txs0:   jnb     S0CON.1,txs0            ;txst0 should not be here
txs1:   clr     S0CON.1 ;debugging only
        mov.b   S0BUF,r1l
        pops
        next

;       T2START         ( c -- )
;       Start Timer 2 and load the count to T2H/T2L and T2CAPH/T2CAPL.

        SCODE   7,'T2START',TSTART
        cpl.w   r1
        clr     TR2
        mov.b   TL2,r1l
        mov.b   TH2,r1h
        mov.b   T2CAPL,r1l
        mov.b   T2CAPH,r1h
        clr     TF2
        setb    TR2
        pops    
        next

;       T2DONE  ( -- )
;       Wait until Timer 2 expires.

        SCODE   6,'T2DONE',TDONE
TDONE1: jnb     TF2,TDONE1
        clr     TF2
        next


;
;       PAUSE   ( -- )
;       Switch task.
;       User areas are independent.  SP, TOS, IP are in its own register
;       group.  Only RP needs saving and UP needs switching.
;       Task0 uses 0-FF as user area and register group 0.
;       Task1 uses 100-1FF as user area and register group 1.
;
        SCODE   5,'PAUSE',PAUSE
        mov.w   r2,r7   ;save RP
        xor.b   PSWH,#10H               ;switch register group
        xor.w   r6,#100H                ;switch UP
        mov.w   r7,r2   ;restore RP of target task
        next    ;continue target task
        

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Common functions rewritten in code
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       UP      ( -- a )
;       Pointer to the user area.

        HEADER  2,'UP',UP
        DW      DOCON
        DW      0

;       ?DUP    ( w -- w w | 0 )
;       Dup tos if it is not zero.

        SCODE   4,'?DUP',QDUP
        cmp.w   r1,#0
        beq     qdup1
        pushs
qdup1:  next

;       ROT     ( w1 w2 w3 -- w2 w3 w1 )
;       Rot 3rd item to top.

        SCODE   3,'ROT',ROT
        xch.w   r1,[r0]
        adds.w  r0,#2
        xch.w   r1,[r0]
        adds.w  r0,#-2
        next
        
;       2DROP   ( w w -- )
;       Discard two items on stack.

        SCODE   5,'2DROP',DDROP
        adds.w  r0,#2
        pops
        next

;       2DUP    ( w1 w2 -- w1 w2 w1 w2 )
;       Duplicate top two items.

        SCODE   4,'2DUP',DDUP
        mov.w   r4,[r0]
        adds.w  r0,#-4
        mov.w   [r0],r4
        mov.w   [r0+2],r1
        next

;       +       ( w w -- sum )
;       Add top two items.

        SCODE   1,'+',PLUS
        add.w   r1,[r0+]
        next

;       D+      ( d d -- d )
;       Double addition.

        SCODE   2,'D+',DPLUS
        mov.w   r4,[r0+]
        add.w   [r0+2],r4
        addc.w  r1,[r0+]
        next

;       D-      ( d d -- d )
;       Double subtraction.

        SCODE   2,'D!-',DMINUS
        mov.w   r4,[r0+]
        sub.w   [r0+2],r4
        subb.w  r1,[r0+]
        neg.w   r1
        next

;       NOT     ( w -- w )
;       One's complement of tos.

        SCODE   3,'NOT',INVER
        cpl.w   r1
        next

;       NEGATE  ( n -- -n )
;       Two's complement of tos.

        SCODE   6,'NEGATE',NEGAT
        neg.w   r1
        next

;       NOOP    ( -- )
;       No operation.

        SCODE   4,'NOOP',NOOP
        next

;       DNEGATE ( d -- -d )
;       Two's complement of top double.

        SCODE   7,'DNEGATE',DNEGA
        movs.w  r4,#0
        movs.w  r5,#0
        sub.w   r5,[r0] 
        subb.w  r4,r1
        mov.w   [r0],r5
        mov.w   r1,r4
        next

;       -       ( n1 n2 -- n1-n2 )
;       Subtraction.

        SCODE   1,'!-',SBBB
        sub.w   r1,[r0+]
        neg.w   r1
        next

;       ABS     ( n -- n )
;       Return the absolute value of n.

        SCODE   3,'ABS',ABSS
        mov.w   r1,r1
        bpl     abs1
        neg.w   r1
abs1:   next

;; Comparison

;       =       ( w w -- t )
;       Return true if top two are equal.

        SCODE   1,'=',EQUAL
        cmp.w   [r0+],r1
        beq     yes
        br      no

;       >       ( n1 n2 -- t )
;       Signed compare of top two items.

        SCODE   1,'>',GREAT
        cmp.w   [r0+],r1
        bgt     yes
        br      no

;       U<      ( u u -- t )
;       Unsigned compare of top two items.

        SCODE   2,'U!<',ULESS
        cmp.w   [r0+],r1
        beq     no
        bl      yes
no:     mov.w   r1,#0
        next
yes:    mov.w   r1,#-1
        next

;       <       ( n1 n2 -- t )
;       Signed compare of top two items.

        SCODE   1,'!<',LESS
        cmp.w   [r0+],r1
        blt     yes
        br      no

;       WITHIN  ( u ul uh -- t )
;       Return true if u is within the range of ul and uh.

        SCODE   6,'WITHIN',WITHI
        mov.w   r4,[r0+]
        mov.w   r5,[r0+]
        cmp.w   r5,r4
        blt     no
        cmp.w   r5,r1
        bgt     no
        br      yes
                
;       MAX     ( n n -- n )
;       Return the greater of two top stack items.

        SCODE   3,'MAX',MAX
        cmp.w   [r0+],r1
        ble     max2
max1:   mov.w   r1,[r0-2]
        align.w
max2:   next
        
;       MIN     ( n n -- n )
;       Return the smaller of top two stack items.

        SCODE   3,'MIN',MIN
        cmp.w   [r0+],r1
        blt     max1
        next
        
;; Divide

;       UM/MOD  ( udl udh u -- ur uq )
;       Unsigned divide of a double by a single. Return mod and quotient.

        SCODE   6,'UM/MOD',UMMOD
        mov.w   r5,[r0+]
        mov.w   r4,[r0]
        divu.d  r4,r1
        mov.w   [r0],r5
        mov.w   r1,r4
        next

;       M/MOD   ( d n -- r q )
;       Signed floored divide of double by single. Return mod and quotient.

        SCODE   5,'M/MOD',MSMOD
        mov.w   r5,[r0+]
        mov.w   r4,[r0]
        div.d   r4,r1
        mov.w   [r0],r5
        mov.w   r1,r4
        next

;       /MOD    ( n n -- r q )
;       Signed divide. Return mod and quotient.

        SCODE   4,'/MOD',SLMOD
        mov.w   r4,[r0]
        sext.w  r5
        div.d   r4,r1
        mov.w   [r0],r5
        mov.w   r1,r4
        next

;       MOD     ( n n -- r )
;       Signed divide. Return mod only.

        SCODE   3,'MOD',MODD
        mov.w   r4,[r0+]
        sext.w  r5
        div.d   r4,r1
        mov.w   r1,r5
        next

;       /       ( n n -- q )
;       Signed divide. Return quotient only.

        SCODE   1,'/',SLASH
        mov.w   r4,[r0+]
        sext.w  r5
        div.d   r4,r1
        mov.w   r1,r4
        next

;       SQRT ( ud -- u )
;       Take square root of an unsigned double integer.

        SCODE   4,'SQRT',SQRT
        push.w  r6,r3,r2,r0             ;save 4 registers I'll use
        mov.w   r0,[r0] ;source in r1:r0
        movs.w  r2,#0   ;init remainder
        movs.w  r3,#0
        movs.w  r4,#0   ;init root
        movs.w  r5,#0
        mov.w   r6,#16  ;loop count
sqrt1:  asl.d   r0,#1   ;shift source 1 bit
        rlc.w   r2,#1   ;shift 1st bit to remainder
        rlc.w   r3,#1
        asl.d   r0,#1   ;shift source again
        rlc.w   r2,#1   ;shift 2nd bits to remainder
        rlc.w   r3,#1
        asl.d   r4,#2   ;shift root and add 1
        adds.w  r4,#1
        cmp.w   r3,r5   ;compare root to accum
        beq     sq0
        bg      sqg     ;* br branch only when greater
        bl      sql     ;* not as defined in databook
sq0:    cmp.w   r2,r4   ;* branch when greater or equal
        bg      sqg     ;* the equal branch must be
        beq     sqg     ;* coded explicitly 15jan98cht
sql:    asr.d   r4,#1   ;acc<root, take out the 1
        djnz    r6,sqrt1
        br      sqrt2
        nop
sqg:    sub.w   r2,r4   ;acc>=root, subtract
        subb.w  r3,r5
        asr.d   r4,#1   ;correct root
        add.w   r4,#1   ;add next bit
        addc.w  r5,#0
        djnz    r6,sqrt1
        nop
sqrt2:  mov.w   r1,r4   ;return root
        pop.w   r0,r2,r3,r6             ;restore registers
        adds.w  r0,#2   ;pop source
        next    


;; Multiply

;       UM*     ( u u -- ud )
;       Unsigned multiply. Return double product.

        SCODE   3,'UM*',UMSTA
        mov.w   r4,[r0]
        mulu.w  r4,r1
        mov.w   [r0],r4
        mov.w   r1,r5
        next

;       *       ( n n -- n )
;       Signed multiply. Return single product.

        SCODE   1,'*',STAR
        mov.w   r4,[r0+]
        mul.w   r4,r1
        mov.w   r1,r4
        next

;       M*      ( n n -- d )
;       Signed multiply. Return double product.

        SCODE   2,'M*',MSTAR
        mov.w   r4,[r0]
        mul.w   r4,r1
        mov.w   [r0],r4
        mov.w   r1,r5
        next

;       */MOD   ( n1 n2 n3 -- r q )
;       Multiply n1 and n2, then divide by n3. Return mod and quotient.

        SCODE   5,'*/MOD',SSMOD
        mov.w   r4,[r0+]
        mov.w   r5,[r0]
        mul.w   r4,r5
        div.d   r4,r1
        mov.w   [r0],r5
        mov.w   r1,r4
        next

;       */      ( n1 n2 n3 -- q )
;       Multiply n1 by n2, then divide by n3. Return quotient only.

        SCODE   2,'*/',STASL
        mov.w   r4,[r0+]
        mov.w   r5,[r0+]
        mul.w   r4,r5
        div.d   r4,r1
        mov.w   r1,r4
        next

;; Miscellaneous

;       CELL+   ( a -- a )
;       Add cell size in byte to address.

        SCODE   5,'CELL+',CELLP
        adds.w  r1,#2
        next

;       CELL-   ( a -- a )
;       Subtract cell size in byte from address.

        SCODE   5,'CELL!-',CELLM
        adds.w  r1,#-2
        next

;       CELLS   ( n -- n )
;       Multiply tos by cell size in bytes.

        SCODE   5,'CELLS',CELLS
        add.w   r1,r1
        next

;       ALIGNED ( b -- a )
;       Align address to the cell boundary.

        SCODE   7,'ALIGNED',ALGND
        adds.w  r1,#1
        and.w   r1,#-2
        next

;       BL      ( -- 32 )
;       Return 32, the blank character.

        SCODE   2,'BL',BLANK
        pushs
        mov.w   r1,#32
        next

;       >CHAR   ( c -- c )
;       Filter non-printing characters.

        SCODE   5,'>CHAR',TCHAR
        and.w   r1,#07fh
        cmp.w   r1,#32
        blt     tcha1
        cmp.w   r1,#127
        blt     tcha2
tcha1:  mov.w   r1,#95
tcha2:  next

;       PICK    ( ... +n -- ... w )
;       Copy the nth stack item to tos.

        SCODE   4,'PICK',PICK
        add.w   r1,r1
        add.w   r1,r0
        mov.w   r1,[r1]
        next

;       ROLL    ( ... +n -- ... )
;       Copy the nth stack item to tos.

        SCODE   4,'ROLL',ROLL
        mov.w   r4,r1
        beq     roll2
        pops
        mov.w   r5,r0
roll1:  xch.w   r1,[r5] ;exchange tos with stack items
        adds.w  r5,#2   ;bump pointer
        djnz    r4,roll1                ;decrement count until done
        next
        align.w
roll2:  pops
        next

;; Memory access

;       +!      ( n a -- )
;       Add n to the contents at address a.

        SCODE   2,'+!!',PSTOR
        mov.w   r4,[r0+]
        add.w   [r1],r4
        pops
        next

;       2!      ( d a -- )
;       Store the double integer to address a.

        SCODE   2,'2!!',DSTOR
        mov.w   r4,[r0+]
        mov.w   r5,[r0+]
        mov.w   [r1+],r4
        mov.w   [r1],r5
        pops
        next

;       2@      ( a -- d )
;       Fetch double integer from address a.

        SCODE   2,'2@',DAT
        mov.w   r4,[r1+]
        mov.w   r5,[r1]
        adds.w  r0,#-2
        mov.w   [r0],r5
        mov.w   r1,r4
        next

;       COUNT   ( b -- b +n )
;       Return count byte of a string and add 1 to byte address.

        SCODE   5,'COUNT',COUNT
        mov.b   r4l,[r1+]
        movs.b  r4h,#0
        pushs
        mov.w   r1,r4
        next

;       @EXECUTE        ( a -- )
;       Execute vector stored in address a.

        SCODE   8,'@EXECUTE',ATEXE
        mov.w   r2,[r1]
        beq     exe1
        pops
        jmp     [[r2+]]
exe1:   pops
        next

; Moves

;       MOVE    ( b1 b2 u -- )
;       Copy u bytes from b1 to b2.

        SCODE   4,'MOVE',MOVE
        mov.w   r4,[r0]
        cmp.w   r4,[r0+2]
        bg      CMOVF+2
        br      CMOVE+2

;       CMOVE   ( b1 b2 u -- )
;       Copy u bytes from b1 to b2.

        SCODE   5,'CMOVE',CMOVE
        mov.w   r4,[r0+]
        mov.w   r5,[r0+]
        cmp.w   r1,#0
        ble     cmov2
cmov1:  mov.b   [r4+],[r5+]
        djnz    r1,cmov1
        nop
cmov2:  pops
        next

;       CMOVE>  ( b1 b2 u -- )
;       Copy u bytes from b1 to b2.

        SCODE   6,'CMOVE>',CMOVF
        mov.w   r4,[r0+]
        mov.w   r5,[r0+]
        cmp.w   r1,#0
        ble     cmvf2
        add.w   r4,r1
        add.w   r5,r1
cmvf1:  adds.w  r4,#-1
        adds.w  r5,#-1
        mov.b   [r4+],[r5+]
        djnz    r1,cmvf1
        nop
cmvf2:  pops
        next

;       FILL    ( b u c -- )
;       Fill u bytes of character c to area beginning at b.

        SCODE   4,'FILL',FILL
        mov.w   r4,[r0+]
        mov.w   r5,[r0+]
fill1:  mov.b   [r5+],r1l
        djnz    r4,fill1
        pops
        next

;       ERASE   ( b u -- )
;       Zero u bytes c to area beginning at b.

        SCODE   5,'ERASE',ERASE
        mov.w   r4,[r0+]
eras1:  mov.b   [r4+],#0
        djnz    r1,eras1
        pops
        next

;       -TRAILING       ( b u -- b u )
;       Adjust the count to eliminate trailing white space.

        SCODE   9,'!-TRAILING',DTRAI
        mov.w   r1,r1
        ble     dtra2
        mov.w   r4,[r0]
        add.w   r4,r1
dtra1:  adds.w  r4,#-1
        cmp.b   [r4],#32
        bgt     dtra2
        djnz    r1,dtra1
        align.w
dtra2:  next


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;User variables to be copied from code space to data space.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

;       ORG     CODEE   ;start code dictionary

;; System and user variables

;       SP0     ( -- a )
;       Pointer to bottom of the data stack.

        SUSER   3,'sp0',SZERO

;       RP0     ( -- a )
;       Pointer to bottom of the return stack.

        SUSER   3,'rp0',RZERO

;       '?KEY   ( -- a )
;       Execution vector of ?KEY.

        SUSER   5,"'?key",TQKEY

;       'EMIT   ( -- a )
;       Execution vector of EMIT.

        SUSER   5,"'emit",TEMIT

;       'EXPECT ( -- a )
;       Execution vector of EXPECT.

        SUSER   7,"'expect",TEXPE

;       'TAP    ( -- a )
;       Execution vector of TAP.

        SUSER   4,"'tap",TTAP

;       'ECHO   ( -- a )
;       Execution vector of ECHO.

        SUSER   5,"'echo",TECHO

;       'PROMPT ( -- a )
;       Execution vector of PROMPT.

        SUSER   7,"'prompt",TPROM

;       BASE    ( -- a )
;       Storage of the radix base for numeric I/O.

        SUSER   4,'BASE',BASE

;       tmp     ( -- a )
;       A temporary storage location used in parse and find.

        SUSER   043H,'tmp',TEMP

;       SPAN    ( -- a )
;       Hold character count received by EXPECT.

        SUSER   4,'SPAN',SPAN

;       >IN     ( -- a )
;       Hold the character pointer while parsing input stream.

        SUSER   3,'>IN',INN

;       #TIB    ( -- a )
;       Hold the current count and address of the terminal input buffer.

        SUSER   4,'#TIB',NTIB
        _USER   SET     _USER+CELLL

;       CSP     ( -- a )
;       Hold the stack pointer for error checking.

        SUSER   3,'csp',CSP

;       'EVAL   ( -- a )
;       Execution vector of EVAL.

        SUSER   5,"'eval",TEVAL

;       'NUMBER ( -- a )
;       Execution vector of NUMBER?.

        SUSER   7,"'number",TNUMB

;       HLD     ( -- a )
;       Hold a pointer in building a numeric output string.

        SUSER   3,'hld',HLD

;       HANDLER ( -- a )
;       Hold the return stack pointer for error handling.

        SUSER   7,'handler',HANDL

;       CONTEXT ( -- a )
;       A area to specify vocabulary search order.

        SUSER   7,'CONTEXT',CNTXT
        _USER   SET     _USER+VOCSS*CELLL       ;vocabulary stack

;       CURRENT ( -- a )
;       Point to the vocabulary to be extended.

        SUSER   7,'CURRENT',CRRNT
        _USER SET       _USER+CELLL             ;vocabulary link pointer

;       CP      ( -- a )
;       Point to the top of the code dictionary.

        SUSER   2,'CP',CP

;       NP      ( -- a )
;       Point to the bottom of the name dictionary.

        SUSER   2,'NP',NP

;       LAST    ( -- a )
;       Point to the last name in the name dictionary.

        SUSER   4,'LAST',LAST

;       forth   ( -- a )
;       Point to the last name in the name dictionary.

        SUSER   5,'forth',VFRTH


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; High level words to be copied to data space
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       FORTH   ( -- )
;       Make FORTH the context vocabulary.

        SCOLON  5,'FORTH',FORTH
        DW      VFRTH,CNTXT,STORE,EXIT

;       DEPTH   ( -- n )
;       Return the depth of the data stack.

        SCOLON  5,'DEPTH',DEPTH
        DW      SPAT,SZERO,AT,SWAP,SBBB,CELLM
        DW      DOLIT,CELLL,SLASH,EXIT

;       HERE    ( -- a )
;       Return the top of the code dictionary.

        SCOLON  4,'HERE',HERE
        DW      CP,AT,EXIT

;       PAD     ( -- a )
;       Return the address of a temporary buffer.

        SCOLON  3,'PAD',PAD
        DW      HERE,DOLIT,80,PLUS,EXIT

;       TIB     ( -- a )
;       Return the address of the terminal input buffer.

        SCOLON  3,'TIB',TIB
        DW      NTIB,CELLP,AT,EXIT

;       PACK$   ( b u a -- a )
;       Build a counted string with u characters from b. Null fill.

        SCOLON  5,'pack$',PACKS
        DW      ALGND,DUPP,TOR          ;strings only on cell boundary
        DW      OVER,DUPP,DOLIT,0
        DW      DOLIT,CELLL,UMMOD,DROP  ;count mod cell
        DW      SBBB,OVER,PLUS
        DW      DOLIT,0,SWAP,STORE      ;null fill cell
        DW      DDUP,CSTOR,DOLIT,1,PLUS ;save count
        DW      SWAP,CMOVE,RFROM,EXIT   ;move string

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Numeric output, single precision
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       DIGIT   ( u -- c )
;       Convert digit u to a character.

        SCOLON  5,'digit',DIGIT
        DW      DOLIT,9,OVER,LESS
        DW      DOLIT,7,ANDD,PLUS
        DW      DOLIT,'0',PLUS,EXIT

;       EXTRACT ( n base -- n c )
;       Extract the least significant digit from n.

        SCOLON  7,'extract',EXTRC
        DW      DOLIT,0,SWAP,UMMOD
        DW      SWAP,DIGIT,EXIT

;       <#      ( -- )
;       Initiate the numeric output process.

        SCOLON  2,'>#',BDIGS
        DW      PAD,HLD,STORE,EXIT

;       HOLD    ( c -- )
;       Insert a character into the numeric output string.

        SCOLON  4,'HOLD',HOLD
        DW      HLD,AT,DOLIT,1,SBBB
        DW      DUPP,HLD,STORE,CSTOR,EXIT

;       #       ( u -- u )
;       Extract one digit from u and append the digit to output string.

        SCOLON  1,'#',DIG
        DW      BASE,AT,EXTRC,HOLD,EXIT

;       #S      ( u -- 0 )
;       Convert u until all digits are added to the output string.

        SCOLON  2,'#S',DIGS
DIGS1:  DW      DIG,DUPP
        DW      QBRAN,DIGS2
        DW      BRAN,DIGS1
DIGS2:  DW      EXIT

;       SIGN    ( n -- )
;       Add a minus sign to the numeric output string.

        SCOLON  4,'SIGN',SIGN
        DW      ZLESS
        DW      QBRAN,SIGN1
        DW      DOLIT,'-',HOLD
SIGN1:  DW      EXIT

;       #>      ( w -- b u )
;       Prepare the output string to be TYPE'd.

        SCOLON  2,'#>',EDIGS
        DW      DROP,HLD,AT
        DW      PAD,OVER,SBBB,EXIT

;       str     ( n -- b u )
;       Convert a signed integer to a numeric string.

        SCOLON  3,'str',STR
        DW      DUPP,TOR,ABSS
        DW      BDIGS,DIGS,RFROM
        DW      SIGN,EDIGS,EXIT

;       HEX     ( -- )
;       Use radix 16 as base for numeric conversions.

        SCOLON  3,'HEX',HEX
        DW      DOLIT,16,BASE,STORE,EXIT

;       DECIMAL ( -- )
;       Use radix 10 as base for numeric conversions.

        SCOLON  7,'DECIMAL',DECIM
        DW      DOLIT,10,BASE,STORE,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Numeric input, single precision
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       DIGIT?  ( c base -- u t )
;       Convert a character to its numeric value. A flag indicates success.

        SCOLON  6,'digit?',DIGTQ
        DW      TOR,DOLIT,'0',SBBB
        DW      DOLIT,9,OVER,LESS
        DW      QBRAN,DGTQ1
        DW      DOLIT,7,SBBB
        DW      DUPP,DOLIT,10,LESS,ORR
DGTQ1:  DW      DUPP,RFROM,ULESS,EXIT

;       NUMBER? ( a -- n T | a F )
;       Convert a number string to integer. Push a flag on tos.

        SCOLON  7,'number?',NUMBQ
        DW      BASE,AT,TOR,DOLIT,0,OVER,COUNT
        DW      OVER,CAT,DOLIT,'$',EQUAL
        DW      QBRAN,NUMQ1
        DW      HEX,SWAP,DOLIT,1,PLUS
        DW      SWAP,DOLIT,1,SBBB
NUMQ1:  DW      OVER,CAT,DOLIT,'-',EQUAL,TOR
        DW      SWAP,RAT,SBBB,SWAP,RAT,PLUS,QDUP
        DW      QBRAN,NUMQ6
        DW      DOLIT,1,SBBB,TOR
NUMQ2:  DW      DUPP,TOR,CAT,BASE,AT,DIGTQ
        DW      QBRAN,NUMQ4
        DW      SWAP,BASE,AT,STAR,PLUS,RFROM
        DW      DOLIT,1,PLUS
        DW      DONXT,NUMQ2
        DW      RAT,SWAP,DROP
        DW      QBRAN,NUMQ3
        DW      NEGAT
NUMQ3:  DW      SWAP
        DW      BRAN,NUMQ5
NUMQ4:  DW      RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:  DW      DUPP
NUMQ6:  DW      RFROM,DDROP
        DW      RFROM,BASE,STORE,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Basic I/O
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        
;       ?KEY    ( -- c T | F )
;       Return input character and true, or a false if no input.

        SCOLON  4,'?KEY',QKEY
        DW      PAUSE
        DW      TQKEY,ATEXE,EXIT

;       KEY     ( -- c )
;       Wait for and return an input character.

        SCOLON  3,'KEY',KEY
KEY1:   
        DW      QKEY
        DW      QBRAN,KEY1
        DW      EXIT

;       EMIT    ( c -- )
;       Send a character to the output device.

        SCOLON  4,'EMIT',EMIT
        DW      TEMIT,ATEXE,EXIT

;       NUF?    ( -- t )
;       Return false if no input, else pause and if CR return true.

        SCOLON  4,'NUF?',NUFQ
        DW      QKEY,DUPP
        DW      QBRAN,NUFQ1
        DW      DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:  DW      EXIT

;       PACE    ( -- )
;       Send a pace character for the file downloading process.

        SCOLON  4,'PACE',PACE
        DW      DOLIT,11,EMIT,EXIT

;       SPACE   ( -- )
;       Send the blank character to the output device.

        SCOLON  5,'SPACE',SPACE
        DW      BLANK,EMIT,EXIT

;       SPACES  ( +n -- )
;       Send n spaces to the output device.

        SCOLON  6,'SPACES',SPACS
        DW      DOLIT,0,MAX,TOR
        DW      BRAN,CHAR2
CHAR1:  DW      SPACE
CHAR2:  DW      DONXT,CHAR1
        DW      EXIT

;       TYPE    ( b u -- )
;       Output u characters from b.

        SCOLON  4,'TYPE',TYPEE
        DW      TOR
        DW      BRAN,TYPE2
TYPE1:  DW      DUPP,CAT,EMIT
        DW      DOLIT,1,PLUS
TYPE2:  DW      DONXT,TYPE1
        DW      DROP,EXIT

;       CR      ( -- )
;       Output a carriage return and a line feed.

        SCOLON  2,'CR',CR
        DW      DOLIT,CRR,EMIT
        DW      DOLIT,LF,EMIT,EXIT

;       $"|     ( -- a )
;       Run time routine compiled by $". Return address of a compiled string.

        SCOLON  043H,'$"|',STRQP
        DW      DOSTR,EXIT              ;force a call to do$

;       ."|     ( -- )
;       Run time routine of ." . Output a compiled string.

        SCOLON  043H,'."|',DOTQP
        DW      DOSTR,COUNT,TYPEE,EXIT

;       .R      ( n +n -- )
;       Display an integer in a field of n columns, right justified.

        SCOLON  2,'.R',DOTR
        DW      TOR,STR,RFROM,OVER,SBBB
        DW      SPACS,TYPEE,EXIT

;       U.R     ( u +n -- )
;       Display an unsigned integer in n column, right justified.

        SCOLON  3,'U.R',UDOTR
        DW      TOR,BDIGS,DIGS,EDIGS
        DW      RFROM,OVER,SBBB
        DW      SPACS,TYPEE,EXIT

;       U.      ( u -- )
;       Display an unsigned integer in free format.

        SCOLON  2,'U.',UDOT
        DW      BDIGS,DIGS,EDIGS
        DW      SPACE,TYPEE,EXIT

;       .       ( w -- )
;       Display an integer in free format, preceeded by a space.

        SCOLON  1,'.',DOT
        DW      BASE,AT,DOLIT,10,XORR   ;?decimal
        DW      QBRAN,DOT1
        DW      UDOT,EXIT               ;no, display unsigned
DOT1:   DW      STR,SPACE,TYPEE,EXIT    ;yes, display signed

;       ?       ( a -- )
;       Display the contents in a memory cell.

        SCOLON  1,'?',QUEST
        DW      AT,DOT,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Parsing
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       parse   ( b u c -- b u delta ; <string> )
;       Scan string delimited by c. Return found string and its offset.

        SCOLON  5,'parse',PARS
        DW      TEMP,STORE,OVER,TOR,DUPP
        DW      QBRAN,PARS8
        DW      DOLIT,1,SBBB,TEMP,AT,BLANK,EQUAL
        DW      QBRAN,PARS3
        DW      TOR
PARS1:  DW      BLANK,OVER,CAT          ;skip leading blanks ONLY
        DW      SBBB,ZLESS,INVER
        DW      QBRAN,PARS2
        DW      DOLIT,1,PLUS
        DW      DONXT,PARS1
        DW      RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:  DW      RFROM
PARS3:  DW      OVER,SWAP
        DW      TOR
PARS4:  DW      TEMP,AT,OVER,CAT,SBBB   ;scan for delimiter
        DW      TEMP,AT,BLANK,EQUAL
        DW      QBRAN,PARS5
        DW      ZLESS
PARS5:  DW      QBRAN,PARS6
        DW      DOLIT,1,PLUS
        DW      DONXT,PARS4
        DW      DUPP,TOR
        DW      BRAN,PARS7
PARS6:  DW      RFROM,DROP,DUPP
        DW      DOLIT,1,PLUS,TOR
PARS7:  DW      OVER,SBBB
        DW      RFROM,RFROM,SBBB,EXIT
PARS8:  DW      OVER,RFROM,SBBB,EXIT

;       PARSE   ( c -- b u ; <string> )
;       Scan input stream and return counted string delimited by c.

        SCOLON  5,'PARSE',PARSE
        DW      TOR,TIB,INN,AT,PLUS     ;current input buffer pointer
        DW      NTIB,AT,INN,AT,SBBB     ;remaining count
        DW      RFROM,PARS,INN,PSTOR,EXIT

;       .(      ( -- )
;       Output following string up to next ) .

        SCOLON  082H,'.(',DOTPR
        DW      DOLIT,')',PARSE,TYPEE,EXIT

;       (       ( -- )
;       Ignore following string up to next ) . A comment.

        SCOLON  081H,'(',PAREN
        DW      DOLIT,')',PARSE,DDROP,EXIT

;       \       ( -- )
;       Ignore following text till the end of line.

        SCOLON  081H,'\',BKSLA
        DW      NTIB,AT,INN,STORE,EXIT

;       CHAR    ( -- c )
;       Parse next word and return its first character.

        SCOLON  4,'CHAR',CHAR
        DW      BLANK,PARSE,DROP,CAT,EXIT

;       TOKEN   ( -- a ; <string> )
;       Parse a word from input stream and copy it to name dictionary.

        SCOLON  5,'token',TOKEN
        DW      BLANK,PARSE,DOLIT,31,MIN
        DW      NP,AT,OVER,SBBB,CELLM
        DW      PACKS,EXIT

;       WORD    ( c -- a ; <string> )
;       Parse a word from input stream and copy it to code dictionary.

        SCOLON  4,'WORD',WORDD
        DW      PARSE,HERE,PACKS,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Dictionary search
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;       NAME>   ( na -- ca )
;       Return a code address given a name address.

        SCOLON  5,'NAME>',NAMET
        DW      CELLM,CELLM,AT,EXIT

;       SAME?   ( a a u -- a a f \ -0+ )
;       Compare u cells in two strings. Return 0 if identical.

        SCOLON  5,'same?',SAMEQ
        DW      TOR
        DW      BRAN,SAME2
SAME1:  DW      OVER,RAT,CELLS,PLUS,AT
        DW      OVER,RAT,CELLS,PLUS,AT
        DW      SBBB,QDUP
        DW      QBRAN,SAME2
        DW      RFROM,DROP,EXIT         ;strings not equal
SAME2:  DW      DONXT,SAME1
        DW      DOLIT,0,EXIT            ;strings equal

;       find    ( a va -- ca na | a F )
;       Search a vocabulary for a string. Return ca and na if succeeded.

        SCOLON  4,'find',FIND
        DW      SWAP,DUPP,CAT
        DW      DOLIT,CELLL,SLASH,TEMP,STORE
        DW      DUPP,AT,TOR,CELLP,SWAP
FIND1:  DW      AT,DUPP
        DW      QBRAN,FIND6
        DW      DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
        DW      QBRAN,FIND2
        DW      CELLP,DOLIT,-1          ;true flag
        DW      BRAN,FIND3
FIND2:  DW      CELLP,TEMP,AT,SAMEQ
FIND3:  DW      BRAN,FIND4
FIND6:  DW      RFROM,DROP
        DW      SWAP,CELLM,SWAP,EXIT
FIND4:  DW      QBRAN,FIND5
        DW      CELLM,CELLM
        DW      BRAN,FIND1
FIND5:  DW      RFROM,DROP,SWAP,DROP
        DW      CELLM
        DW      DUPP,NAMET,SWAP,EXIT

;       NAME?   ( a -- ca na | a F )
;       Search all context vocabularies for a string.

        SCOLON  5,'name?',NAMEQ
        DW      CNTXT,DUPP,DAT,XORR     ;?context=also
        DW      QBRAN,NAMQ1
        DW      CELLM   ;no, start with context
NAMQ1:  DW      TOR
NAMQ2:  DW      RFROM,CELLP,DUPP,TOR    ;next in search order
        DW      AT,QDUP
        DW      QBRAN,NAMQ3
        DW      FIND,QDUP               ;search vocabulary
        DW      QBRAN,NAMQ2
        DW      RFROM,DROP,EXIT         ;found name
NAMQ3:  DW      RFROM,DROP              ;name not found
        DW      DOLIT,0,EXIT            ;false flag

;; Terminal response

;       ^H      ( bot eot cur -- bot eot cur )
;       Backup the cursor by one character.

        SCOLON  2,'^h',BKSP
        DW      TOR,OVER,RFROM,SWAP,OVER,XORR
        DW      QBRAN,BACK1
        DW      DOLIT,BKSPP,TECHO,ATEXE,DOLIT,1,SBBB
        DW      BLANK,TECHO,ATEXE
        DW      DOLIT,BKSPP,TECHO,ATEXE
BACK1:  DW      EXIT

;       TAP     ( bot eot cur c -- bot eot cur )
;	Accept and echo the key stroke and bump the cursor.

	SCOLON	3,'TAP',TAP
	DW	DUPP,TECHO,ATEXE
	DW	OVER,CSTOR,DOLIT,1,PLUS,EXIT

;	kTAP	( bot eot cur c -- bot eot cur )
;	Process a key stroke, CR or backspace.

	SCOLON	4,'kTap',KTAP
	DW	DUPP,DOLIT,CRR,XORR
	DW	QBRAN,KTAP2
	DW	DOLIT,BKSPP,XORR
	DW	QBRAN,KTAP1
	DW	BLANK,TAP,EXIT
KTAP1:	DW	BKSP,EXIT
KTAP2:	DW	DROP,SWAP,DROP,DUPP,EXIT

;	accept	( b u -- b u )
;	Accept characters to input buffer. Return with actual count.

	SCOLON	6,'accept',ACCEP
	DW	OVER,PLUS,OVER
ACCP1:	DW	DDUP,XORR
	DW	QBRAN,ACCP4
	DW	KEY,DUPP
;	DW	BLANK,SBBB,DOLIT,95,ULESS
	DW	BLANK,DOLIT,127,WITHI
	DW	QBRAN,ACCP2
	DW	TAP
	DW	BRAN,ACCP3
ACCP2:	DW	TTAP,ATEXE
ACCP3:	DW	BRAN,ACCP1
ACCP4:	DW	DROP,OVER,SBBB,EXIT

;	EXPECT	( b u -- )
;	Accept input stream and store count in SPAN.

	SCOLON	6,'EXPECT',EXPEC
	DW	TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;	QUERY	( -- )
;	Accept input stream to terminal input buffer.

	SCOLON	5,'QUERY',QUERY
	DW	TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
	DW	DROP,DOLIT,0,INN,STORE,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Error handling
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	CATCH	( ca -- 0 | err# )
;	Execute word at ca and set up an error frame for it.

	SCOLON	5,'CATCH',CATCH
	DW	SPAT,TOR,HANDL,AT,TOR	;save error frame
	DW	RPAT,HANDL,STORE,EXECU	;execute
	DW	RFROM,HANDL,STORE	;restore error frame
	DW	RFROM,DROP,DOLIT,0,EXIT	;no error

;	THROW	( err# -- err# )
;	Reset system to current local error frame an update error flag.

	SCOLON	5,'THROW',THROW
	DW	HANDL,AT,RPSTO		;restore return stack
	DW	RFROM,HANDL,STORE	;restore handler frame
	DW	RFROM,SWAP,TOR,SPSTO	;restore data stack
	DW	DROP,RFROM,EXIT

;	NULL$	( -- a )
;	Return address of a null string with zero count.

	HEADER	5,'null$',NULLS
	DW	DOVAR
	DW	0
	DB	99,111,121,111,116,101	;eForth signature
	ALIGN.W

;	ABORT	( -- )
;	Reset data stack and jump to QUIT.

	SCOLON	5,'ABORT',ABORT
	DW	NULLS,THROW

;	abort"	( f -- )
;	Run time routine of ABORT" . Abort with a message.

	SCOLON	046H,'abort"',ABORQ
	DW	QBRAN,ABOR1		;text flag
	DW	DOSTR,THROW		;pass error string
ABOR1:	DW	DOSTR,DROP,EXIT		;drop error

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; The text interpreter
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	$INTERPRET	( a -- )
;	Interpret a word. If failed, try to convert it to an integer.

	SCOLON	10,'$interpret',INTER
	DW	NAMEQ,QDUP		;?defined
	DW	QBRAN,INTE1
	DW	AT,DOLIT,COMPO,ANDD	;?compile only lexicon bits
	DSS	ABORQ,'compile-only'
	DW	EXECU,EXIT		;execute defined word
INTE1:	DW	TNUMB,ATEXE		;convert a number
	DW	QBRAN,INTE2
	DW	EXIT
INTE2:	DW	THROW	;error

;	[	( -- )
;	Start the text interpreter.

	SCOLON	081H,'[',LBRAC
	DW	DOLIT,INTER,TEVAL,STORE,EXIT

;	.OK	( -- )
;	Display 'ok' only while interpreting.

	SCOLON	3,'.ok',DOTOK
	DW	DOLIT,INTER,TEVAL,AT,EQUAL
	DW	QBRAN,DOTO1
	DSS	DOTQP,'ok'
DOTO1:	DW	CR,EXIT

;	?STACK	( -- )
;	Abort if the data stack underflows.

	SCOLON	6,'?stack',QSTAC
	DW	DEPTH,ZLESS		;check only for underflow
	DSS	ABORQ,'! underflow'
	DW	EXIT

;	EVAL	( -- )
;	Interpret the input stream.

	SCOLON	4,'eval',EVAL
EVAL1:	DW	TOKEN,DUPP,CAT		;?input stream empty
	DW	QBRAN,EVAL2
	DW	TEVAL,ATEXE,QSTAC	;evaluate input, check stack
	DW	BRAN,EVAL1
EVAL2:	DW	DROP,TPROM,ATEXE,EXIT	;prompt

;; Shell

;	PRESET	( -- )
;	Reset data stack pointer and the terminal input buffer.

	SCOLON	6,'preset',PRESE
	DW	SZERO,AT,SPSTO
;	DW	DOLIT,TIBB,NTIB,CELLP,STORE,EXIT
	DW	EXIT


;	QUIT	( -- )
;	Reset return stack pointer and start text interpreter.

	SCOLON	4,'QUIT',QUIT
	DW	RZERO,AT,RPSTO		;reset return stack pointer
QUIT1:	DW	LBRAC	;start interpretation
QUIT2:	DW	QUERY,SPACE		;get input
	DW	DOLIT,EVAL,CATCH,QDUP	;evaluate input
	DW	QBRAN,QUIT2		;continue till error
	DW	TPROM,AT,SWAP		;save input device
	DW	NULLS,OVER,XORR	;?display error message
	DW	QBRAN,QUIT3
	DW	COUNT,TYPEE,SPACE	;error message
	DSS	DOTQP,'?'		;error prompt
QUIT3:	DW	DOLIT,DOTOK,XORR	;?file input
	DW	QBRAN,QUIT4
	DW	DOLIT,ERR,EMIT		;file error, tell host
QUIT4:	DW	CR,PRESE		;some cleanup
	DW	BRAN,QUIT1

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; The compiler
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	'	( -- ca )
;	Search context vocabularies for the next word in input stream.

	SCOLON	1,"'",TICK
	DW	TOKEN,NAMEQ		;?defined
	DW	QBRAN,TICK1
	DW	EXIT	;yes, push code address
TICK1:	DW	THROW	;no, error

;	ALLOT	( n -- )
;	Allocate n bytes to the code dictionary.

	SCOLON	5,'ALLOT',ALLOT
	DW	CP,PSTOR,EXIT		;adjust code pointer

;	,	( w -- )
;	Compile an integer into the code dictionary.

	SCOLON	1,02CH,COMMA
	DW	HERE,DUPP,CELLP		;cell boundary
	DW	CP,STORE,STORE,EXIT	;adjust code pointer, compile

;	[COMPILE]	( -- ; <string> )
;	Compile the next immediate word into code dictionary.

	SCOLON	089H,'[COMPILE]',BCOMP
	DW	TICK,COMMA,EXIT

;	COMPILE	( -- )
;	Compile the next address in colon list to code dictionary.

	SCOLON	047H,'COMPILE',COMPI
	DW	RFROM,DUPP,AT,COMMA	;compile address
	DW	CELLP,TOR,EXIT		;adjust return address

;	LITERAL	( w -- )
;	Compile tos to code dictionary as an integer literal.

	SCOLON	087H,'LITERAL',LITER
	DW	COMPI,DOLIT,COMMA,EXIT

;	$,"	( -- )
;	Compile a literal string up to next " .

	SCOLON	3,'$!,"',STRCQ
	DW	DOLIT,'"',WORDD		;move string to code dictionary
	DW	COUNT,PLUS,ALGND	;calculate aligned end of string
	DW	CP,STORE,EXIT		;adjust the code pointer

;	RECURSE	( -- )
;	Make the current word available for compilation.

	SCOLON	087H,'RECURSE',RECUR
	DW	LAST,AT,NAMET,COMMA,EXIT

;; Structures

;	FOR	( -- a )
;	Start a FOR-NEXT loop structure in a colon definition.

	SCOLON	083H,'FOR',FOR
	DW	COMPI,TOR,HERE,EXIT

;	BEGIN	( -- a )
;	Start an infinite or indefinite loop structure.

	SCOLON	085H,'BEGIN',BEGIN
	DW	HERE,EXIT

;	NEXT	( a -- )
;	Terminate a FOR-NEXT loop structure.

	SCOLON	084H,'NEXT',NEXT
	DW	COMPI,DONXT,COMMA,EXIT

;	UNTIL	( a -- )
;	Terminate a BEGIN-UNTIL indefinite loop structure.

	SCOLON	085H,'UNTIL',UNTIL
	DW	COMPI,QBRAN,COMMA,EXIT

;	AGAIN	( a -- )
;	Terminate a BEGIN-AGAIN infinite loop structure.

	SCOLON	085H,'AGAIN',AGAIN
	DW	COMPI,BRAN,COMMA,EXIT

;	IF	( -- A )
;	Begin a conditional branch structure.

	SCOLON	082H,'IF',IFF
	DW	COMPI,QBRAN,HERE
	DW	DOLIT,0,COMMA,EXIT

;	AHEAD	( -- A )
;	Compile a forward branch instruction.

	SCOLON	085H,'AHEAD',AHEAD
	DW	COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;	REPEAT	( A a -- )
;	Terminate a BEGIN-WHILE-REPEAT indefinite loop.

	SCOLON	086H,'REPEAT',REPEA
	DW	AGAIN,HERE,SWAP,STORE,EXIT

;	THEN	( A -- )
;	Terminate a conditional branch structure.

	SCOLON	084H,'THEN',THENN
	DW	HERE,SWAP,STORE,EXIT

;	AFT	( a -- a A )
;	Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

	SCOLON	083H,'AFT',AFT
	DW	DROP,AHEAD,BEGIN,SWAP,EXIT

;	ELSE	( A -- A )
;	Start the false clause in an IF-ELSE-THEN structure.

	SCOLON	084H,'ELSE',ELSEE
	DW	AHEAD,SWAP,THENN,EXIT

;	WHILE	( a -- A a )
;	Conditional branch out of a BEGIN-WHILE-REPEAT loop.

	SCOLON	085H,'WHILE',WHILE
	DW	IFF,SWAP,EXIT

;	ABORT"	( -- ; <string> )
;	Conditional abort with an error message.

	SCOLON	086H,'ABORT"',ABRTQ
	DW	COMPI,ABORQ,STRCQ,EXIT

;	$"	( -- ; <string> )
;	Compile an inline string literal.

	SCOLON	082H,'$"',STRQ
	DW	COMPI,STRQP,STRCQ,EXIT

;	."	( -- ; <string> )
;	Compile an inline string literal to be typed out at run time.

	SCOLON	082H,'."',DOTQ
	DW	COMPI,DOTQP,STRCQ,EXIT

;; Name compiler

;	?UNIQUE	( a -- a )
;	Display a warning message if the word already exists.

	SCOLON	7,'?unique',UNIQU
	DW	DUPP,NAMEQ		;?name exists
	DW	QBRAN,UNIQ1		;redefinitions are OK
	DSS	DOTQP,'reDef! '		;but warn the user
	DW	OVER,COUNT,TYPEE	;just in case its not planned
UNIQ1:	DW	DROP,SPACE,EXIT

;	$,n	( na -- )
;	Build a new dictionary name using the string at na.

	SCOLON	3,'$!,n',SNAME
	DW	DUPP,CAT		;?null input
	DW	QBRAN,PNAM1
	DW	UNIQU	;?redefinition
	DW	DUPP,LAST,STORE		;save na for vocabulary link
	DW	HERE,ALGND,SWAP		;align code address
	DW	CELLM	;link address
	DW	CRRNT,AT,AT,OVER,STORE
	DW	CELLM,DUPP,NP,STORE	;adjust name pointer
	DW	STORE,EXIT		;save code pointer
PNAM1:	DSS	STRQP,'name'		;null input
	DW	THROW

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; FORTH compiler
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	$COMPILE	( a -- )
;	Compile next word to code dictionary as a token or literal.

	SCOLON	8,'$compile',SCOMP
	DW	NAMEQ,QDUP		;?defined
	DW	QBRAN,SCOM2
	DW	AT,DOLIT,IMEDD,ANDD	;?immediate
	DW	QBRAN,SCOM1
	DW	EXECU,EXIT		;its immediate, execute
SCOM1:	DW	COMMA,EXIT		;its not immediate, compile
SCOM2:	DW	TNUMB,ATEXE		;try to convert to number
	DW	QBRAN,SCOM3
	DW	LITER,EXIT		;compile number as integer
SCOM3:	DW	THROW	;error

;	OVERT	( -- )
;	Link a new word into the current vocabulary.

	SCOLON	5,'OVERT',OVERT
	DW	LAST,AT,CRRNT,AT,STORE,EXIT

;	;	( -- )
;	Terminate a colon definition.

	SCOLON	0C1H,03BH,SEMIS
	DW	COMPI,EXIT,LBRAC,OVERT,EXIT

;	]	( -- )
;	Start compiling the words in the input stream.

	SCOLON	1,']',RBRAC
	DW	DOLIT,SCOMP,TEVAL,STORE,EXIT

;	:	( -- ; <string> )
;	Start a new colon definition using next word as its name.

	SCOLON	1,':',COLON
	DW	TOKEN,SNAME,DOLIT,DOLST
	DW	COMMA,RBRAC,EXIT

;	IMMEDIATE	( -- )
;	Make the last compiled word an immediate word.

	SCOLON	9,'IMMEDIATE',IMMED
	DW	DOLIT,IMEDD,LAST,AT,AT,ORR
	DW	LAST,AT,STORE,EXIT

;; Defining words

;	USER	( u -- ; <string> )
;	Compile a new user variable.

	SCOLON	4,'USER',USER
	DW	TOKEN,SNAME,OVERT
	DW	DOLIT,DOUSE,COMMA
	DW	COMMA,EXIT

;	CONSTANT	( u -- ; <string> )
;	Compile a new constant.

	SCOLON	8,'CONSTANT',CONST
	DW	TOKEN,SNAME,OVERT
	DW	DOLIT,DOCON,COMMA
	DW	COMMA,EXIT

;	CREATE	( -- ; <string> )
;	Compile a new array entry without allocating code space.

	SCOLON	6,'CREATE',CREAT
	DW	TOKEN,SNAME,OVERT
	DW	DOLIT,DOVAR,COMMA
	DW	EXIT

;	VARIABLE	( -- ; <string> )
;	Compile a new variable initialized to 0.

	SCOLON	8,'VARIABLE',VARIA
	DW	CREAT,DOLIT,0,COMMA,EXIT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Tools
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	_TYPE	( b u -- )
;	Display a string. Filter non-printing characters.

	SCOLON	5,'_TYPE',UTYPE
	DW	TOR	;start count down loop
	DW	BRAN,UTYP2		;skip first pass
UTYP1:	DW	DUPP,CAT,TCHAR,EMIT	;display only printable
	DW	DOLIT,1,PLUS		;increment address
UTYP2:	DW	DONXT,UTYP1		;loop till done
	DW	DROP,EXIT

;	dm+	( a u -- a )
;	Dump u bytes from , leaving a+u on the stack.

	SCOLON	3,'dm+',DMP
	DW	OVER,DOLIT,4,UDOTR	;display address
	DW	SPACE,TOR		;start count down loop
	DW	BRAN,PDUM2		;skip first pass
PDUM1:	DW	DUPP,CAT,DOLIT,3,UDOTR	;display numeric data
	DW	DOLIT,1,PLUS		;increment address
PDUM2:	DW	DONXT,PDUM1		;loop till done
	DW	EXIT

;	DUMP	( a u -- )
;	Dump u bytes from a, in a formatted manner.

	SCOLON	4,'DUMP',DUMP
	DW	BASE,AT,TOR,HEX		;save radix, set hex
	DW	DOLIT,0
	DW	DOLIT,16,UMMOD		;change count to lines
	DW	SWAP,DOLIT,0,EQUAL
	DW	PLUS	;fraction of a line
	DW	TOR	;start count down loop
DUMP1:	DW	CR,DOLIT,16,DDUP,DMP	;display numeric
	DW	ROT,ROT
	DW	SPACE,SPACE,UTYPE	;display printable characters
	DW	NUFQ,INVER		;user control
	DW	QBRAN,DUMP2
	DW	DONXT,DUMP1		;loop till done
	DW	BRAN,DUMP3
DUMP2:	DW	RFROM,DROP		;cleanup loop stack, early exit
DUMP3:	DW	DROP,RFROM,BASE,STORE	;restore radix
	DW	EXIT

;	.S	( ... -- ... )
;	Display the contents of the data stack.

	SCOLON	2,'.S',DOTS
	DW	DEPTH	;stack depth
	DW	TOR	;start count down loop
	DW	BRAN,DOTS2		;skip first pass
DOTS1:	DW	RAT,PICK,DOT		;index stack, display contents
DOTS2:	DW	DONXT,DOTS1		;loop till done
	DW	EXIT

;	>NAME	( ca -- na | F )
;	Convert code address to a name address.

	SCOLON	5,'>name',TNAME
	DW	CNTXT,AT		;vocabulary link
;TNAM1:	DW	CELLP,AT,QDUP		;check all vocabularies
;	DW	QBRAN,TNAM4
	DW	DDUP
TNAM2:	DW	AT,DUPP	;?last word in a vocabulary
	DW	QBRAN,TNAM3
	DW	DDUP,NAMET,XORR		;compare
	DW	QBRAN,TNAM3
	DW	CELLM	;continue with next word
	DW	BRAN,TNAM2
TNAM3:	DW	SWAP,DROP,QDUP
	DW	QBRAN,TNAM5
	DW	SWAP,DROP,SWAP,DROP,EXIT
TNAM5:	DW	SWAP,DROP
TNAM4:	DW	DROP,DOLIT,0,EXIT	;false flag

;	.ID	( na -- )
;	Display the name at address.

	SCOLON	3,'.id',DOTID
	DW	QDUP	;if zero no name
	DW	QBRAN,DOTI1
	DW	COUNT,DOLIT,01FH,ANDD	;mask lexicon bits
	DW	UTYPE,EXIT		;display name string
DOTI1:	DSS	DOTQP,'{noName}'
	DW	EXIT

;	SEE	( -- ; <string> )
;	A simple decompiler.

	SCOLON	3,'SEE',SEE
	DW	TICK	;starting address
	DW	CR
SEE1:	DW	CELLP,DUPP,AT,DUPP	;?does it contain a zero
	DW	QBRAN,SEE2
	DW	TNAME	;?is it a name
SEE2:	DW	QDUP	;name address or zero
	DW	QBRAN,SEE3
	DW	SPACE,DOTID		;display name
	DW	BRAN,SEE4
SEE3:	DW	DUPP,AT,UDOT		;display number
SEE4:	DW	NUFQ	;user control
	DW	QBRAN,SEE1
	DW	DROP,EXIT

;	WORDS	( -- )
;	Display the names in the context vocabulary.

	SCOLON	5,'WORDS',WORDS
	DW	CR,CNTXT,AT		;only in context
WORS1:	DW	AT,QDUP	;?at end of list
	DW	QBRAN,WORS2
	DW	DUPP,SPACE,DOTID	;display a name
	DW	CELLM,NUFQ		;user control
	DW	QBRAN,WORS1
	DW	DROP
WORS2:	DW	EXIT

;	READ	( -- )
;	Read a file into file buffer at 8000H.
;	Terminated by ESC, total bytes stored in CSP.

	SCOLON	4,'READ',READ
	DW	DOLIT,8000H,DOLIT,7FFFH
	DW	TOR
READ1:	DW	KEY,DUPP,DOLIT,27,XORR
	DW	QBRAN,READ2
	DW	DUPP,DOLIT,0DH,EQUAL
	DW	QBRAN,READ4
	DW	DOLIT,'.',TXSTO
READ4:	DW	OVER,CSTOR
	DW	DOLIT,1,PLUS
	DW	DONXT,READ1
	DW	DROP
	DW	BRAN,READ3
READ2:	DW	RFROM,DDROP
READ3:	DW	DOLIT,8000H,SBBB
	DW	CSP,STORE,EXIT


;	LOAD	( -- )
;	Load a file in the file buffer at 8000H.
;	Use total bytes stored in CSP.

	SCOLON	4,'LOAD',LOAD
	DW	INN,AT,NTIB,DAT		;save >IN,#TIB,TIB
	DW	TOR,TOR,TOR		
	DW	DOLIT,0,INN,STORE	;init >IN
	DW	DOLIT,8000H,NTIB,CELLP,STORE
	DW	CSP,AT,NTIB,STORE	;init #TIB and TIB
	DW	EVAL	;load source
	DW	RFROM,RFROM,RFROM	;restore >IN,#TIB,TIB
	DW	NTIB,DSTOR,INN,STORE
	DW	EXIT

;	SEND	( -- )
;	Dump range of memory in Intel Hex format, to be written
;	to PROM.

	SCOLON	4,'SEND',SEND
	DW	HEX,TOR
	DW	BRAN,SEND5
SEND1:	DW	DUPP,DOLIT,31,ANDD
	DW	QBRAN,SEND2
	DW	BRAN,SEND3
SEND2:	DW	CR,DUPP,DOLIT,0FFH,ANDD
	DW	CSP,STORE
	DW	DUPP,DOLIT,100H,SLASH
	DW	CSP,PSTOR
	DW	DOLIT,20H,CSP,PSTOR
	DSS	DOTQP,':20'
	DW	DUPP,BDIGS
	DW	DIG,DIG,DIG,DIG
	DW	EDIGS,TYPEE
	DSS	DOTQP,'00'
SEND3:	DW	COUNT,DUPP
	DW	CSP,PSTOR
	DW	BDIGS,DIG,DIG
	DW	EDIGS,TYPEE
	DW	DUPP,DOLIT,31,ANDD
	DW	QBRAN,SEND4
	DW	BRAN,SEND5
SEND4:	DW	CSP,CAT,NEGAT
	DW	BDIGS,DIG,DIG
	DW	EDIGS,TYPEE
SEND5:	DW	DONXT
	DW	SEND1
	DW	CR,DROP
	DSS	DOTQP,':00000001FF'
	DW	CR,EXIT	

;	FORGET	( -- ; <string> )
;	Trim the Forth vocabulary to the next word.
;	A dumb FORGET, no good for multiple vocabulary treads.

	SCOLON	6,'FORGET',FORGET
	DW	TICK	;starting address
	DW	DUPP,CP,STORE
	DW	TNAME,CELLM,AT
	DW	DUPP,CELLM,CELLM
	DW	NP,STORE
	DW	DUPP,LAST,DSTOR
	DW	EXIT


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Software reset
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;	VER	( -- n )
;	Return the version number of this implementation.

	HEADER	3,'ver',VERSN
	DW	DOCON
	DW	VER*256+EXT

;	hi	( -- )
;	Display the sign-on message of eForth.

	SCOLON	2,'hi',HI
	DW	CR	
	DSS	DOTQP,'XAmon-V'		;model
	DW	BASE,AT,HEX		;save radix
	DW	VERSN,BDIGS,DIG,DIG
	DW	DOLIT,'.',HOLD
	DW	DIGS,EDIGS,TYPEE	;format version number
	DW	BASE,STORE,CR,EXIT	;restore radix

;	'BOOT	( -- a )
;	The application startup vector.

	HEADER 	5,"'boot",TBOOT
	DW	DOVAR
	DW	HI	;application to boot

;	COLD	( -- )
;	The high level cold start sequence.

	SCOLON	4,'COLD',COLD
COLD1:	
	DW	PRESE	;initialize stack and TIB
	DW	TBOOT,ATEXE		;application boot
	DW	FORTH,CNTXT,AT,DUPP	;initialize search order
	DW	CRRNT,DSTOR,OVERT
	DW	QUIT	;start interpretation
	DW	BRAN,COLD1		;just in case


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;; Dictionary pointers
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

LASTN	EQU	_NAME+4	;last name address

NTOP	EQU	_NAME-0	;next memory in name dictionary
CTOP	EQU	$+0	;next memory in code dictionary

	END	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

